import doctest


class CodeChunk(object):
    def __init__(self, source_obj, source, want=None, exc_msg=None,
                 lineno=None, indent=0):
        self.source = source
        self.want = want
        self.exc_msg = exc_msg
        self.lineno = lineno
        self.indent = indent
        self.source_obj = source_obj
        self.options = {}

    def __str__(self):
        return self.source

    def __repr__(self):
        return "<YALPT CodeChunk: %s>" % repr(self.source)


class DocTestParser(object):
    def parse(self, literate_string, file_name):
        parser = doctest.DocTestParser()

        for chunk in parser.parse(literate_string, file_name):
            if isinstance(chunk, doctest.Example):
                yield CodeChunk(chunk, chunk.source, chunk.want, chunk.exc_msg,
                                chunk.lineno, chunk.indent)
            else:
                yield chunk


class MarkdownParser(object):
    def parse(self, literate_string, file_name):
        lines = literate_string.splitlines()
        lines.append(None)

        state = None
        block_start = None
        acc = []
        last_indent = None

        for lineno, line in enumerate(lines):
            first = None
            last_state = state

            if line is None:
                # last line
                state = 'last_line'
            elif state in (None, 'blank_line'):
                if line in ('```', '```python'):
                    state = 'fenced_python'
                    last_indent = 0
                    block_start = lineno + 1
                elif line.startswith('```'):
                    state = 'fenced_other'
                    acc.append(line)
                elif line == '':
                    state = 'blank_line'
                    acc.append(line)
                elif state == 'blank_line' and line.startswith('    '):
                    state = 'indented_python'
                    block_start = lineno
                    first = line[4:]
                else:
                    state = None
                    acc.append(line)
            elif state == 'fenced_python':
                if line == '```':
                    state = None
                elif line != '':
                    # we skip blanks unless dropping down an indentation level
                    # so that the interactive console
                    # doesn't complain
                    indent = len(line) - len(line.lstrip())
                    if indent == 0 and last_indent is not None:
                        if last_indent == 0:
                            state = 'mid_fenced'
                            first = line
                        else:
                            acc.append('')
                            first = line
                            state = 'mid_fenced'
                    else:
                        acc.append(line)

                    last_indent = indent
            elif state == 'fenced_other':
                if line == '```':
                    state = None

                acc.append(line)
            else:
                if line.startswith('    '):
                    state = 'indented_python'
                    acc.append(line[4:])
                elif line == '':
                    state = 'blank_line'
                    first = line
                else:
                    state = None
                    first = line

            if last_state != state:
                if (last_state in ('indented_python', 'fenced_python')
                        or state == 'mid_fenced'):
                    raw_chunk = '\n'.join(acc)+'\n'
                    yield CodeChunk(raw_chunk, raw_chunk, lineno=block_start)
                    if first is not None:
                        acc = [first]
                    else:
                        acc = []

                    if state == 'mid_fenced':
                        state = 'fenced_python'
                        block_start = lineno + 1
                elif state in ('indented_python', 'fenced_python'):
                    yield '\n'.join(acc) + '\n'
                    if first is not None:
                        acc = [first]
                    else:
                        acc = []

                    last_indent = None
                elif state == 'last_line':
                    # it's the last line, and we're in a text block
                    yield '\n'.join(acc)



# Copyright 2014, Solly Ross (see LICENSE.txt)
# Portions Copyright Python Software Foundation
# (see COPYRIGHT.txt and PYTHON-LICENSE.txt)

# main part
import code
import contextlib
import doctest
import getpass
import linecache
import pdb
import re
import sys
import traceback
import warnings

import six

from yalpt import ansi_helper as ansi
from yalpt import formatters
from yalpt import parsers


__all__ = ["LiterateInterpreter"]


@contextlib.contextmanager
def noop_mgr(writer):
    yield writer


class LiterateInterpreter(code.InteractiveConsole):
    def __init__(self, text_formatter=formatters.NoopFormatter(),
                 code_parser=parsers.DocTestParser(), use_ansi=True,
                 use_readline=True, env_driver=None, *args, **kwargs):
        code.InteractiveConsole.__init__(self, *args, **kwargs)

        self._output_checker = doctest.OutputChecker()
        self._fakeout = doctest._SpoofOut()
        self.chunks = None
        self.exc_msg = None
        self.name = 'literate program'
        self.text_formatter = text_formatter
        self.code_parser = code_parser
        self.use_ansi = use_ansi
        self.pause = True
        self.interactive = True

        if use_readline:
            self._readline = __import__('readline')
            self._add_readline()
        else:
            self._readline = None

        self._env_driver = env_driver

        self._correct_path()

    def runfunction(self, func):
        self.runcode(six.get_function_code(func))

    def _correct_path(self):
        def correct_path():
            import sys
            if '' not in sys.path:
                sys.path.insert(0, '')
            del sys

        self.runfunction(correct_path)

    def _add_readline(self):
        self.locals['__console_locals__'] = self.locals
        self.locals['readline'] = self._readline

        def add_readline():
            # add support for completion
            import atexit
            import os
            import readline
            import rlcompleter

            completer = rlcompleter.Completer(__console_locals__)  # noqa
            readline.set_completer(completer.complete)

            # tab completion
            readline.parse_and_bind('Control-space: complete')
            # history file
            histfile = os.path.join(os.environ['HOME'],
                                    '.literate-python-history')
            try:
                readline.read_history_file(histfile)
            except IOError:
                pass
            atexit.register(readline.write_history_file, histfile)
            del os, histfile, readline, rlcompleter

        self.runfunction(add_readline)

        del self.locals['__console_locals__']
        del self.locals['readline']

    def _interact_once(self, more):
        try:
            if more:
                prompt = sys.ps2
            else:
                prompt = sys.ps1
            try:
                line = self.raw_input(prompt)
                # Can be None if sys.stdin was redefined
                encoding = getattr(sys.stdin, "encoding", None)
                if encoding and not isinstance(line, six.text_type):
                    line = line.decode(encoding)
            except EOFError:
                self.write("\n")
                return (False, None)
            else:
                if len(line) == 0:
                    blank = True
                else:
                    blank = False
                more = self.push(line)
        except KeyboardInterrupt:
            self.write("\nKeyboardInterrupt\n")
            self.resetbuffer()
            more = False
            blank = False

        return (blank, more)

    # BEGIN FROM PYTHON STD LIB
    # (extracted from the doctest module, made slight changes to the
    # filename RE, and refactored _capture_output into its own method)
    __LINECACHE_FILENAME_RE = re.compile(r'<literate '
                                         r'(?P<name>.+)'
                                         r'\[(?P<chunknum>\d+)\]>$')

    def _patched_linecache_getlines(self, filename, module_globals=None):
        m = self.__LINECACHE_FILENAME_RE.match(filename)
        if m and m.group('name') == self.name:
            chunk = self.chunks[int(m.group('chunknum'))]
            source = chunk.source
            if isinstance(source, six.text_type):
                source = source.encode('ascii', 'backslashreplace')
            return source.splitlines(True)
        else:
            return self.save_linecache_getlines(filename, module_globals)

    @contextlib.contextmanager
    def _capture_output(self):
        save_stdout = sys.stdout
        sys.stdout = self._fakeout

        # set up pdb to work properly
        save_set_trace = pdb.set_trace
        self.debugger = doctest._OutputRedirectingPdb(save_stdout)
        self.debugger.reset()
        pdb.set_trace = self.debugger.set_trace
        self.save_linecache_getlines = linecache.getlines
        linecache.getlines = self._patched_linecache_getlines
        save_displayhook = sys.displayhook
        sys.displayhook = sys.__displayhook__

        try:
            yield sys.stdout
        finally:
            sys.stdout = save_stdout
            pdb.set_trace = save_set_trace
            linecache.getlines = self.save_linecache_getlines
            sys.displayhook = save_displayhook
    # END FROM PYTHON STD LIB

    def _process_code_line(self, line, res, more):
        if self._readline is not None:
            self._readline.add_history(line)

        if more:
            self.write(sys.ps2)
        else:
            self.write(sys.ps1)

        if self.use_ansi:
            self.write(ansi.with_codes(line, 38, 5, 0xdf))
        else:
            self.write(line)

        self.write("\n")
        with self._capture_output() as output:
            more = self.push(line)

            res += output.getvalue()
            output.truncate(0)

        exc = self.exc_msg
        self.exc_msg = None
        return (res, exc, more)

    def _format_tb(self):
        try:
            extype, value, tb = sys.exc_info()
            sys.last_type = extype
            sys.last_value = value
            sys.last_traceback = tb
            res = traceback.format_exception_only(extype, value)
        finally:
            tb = None

        if res:
            return res[-1]

    def showtraceback(self):
        try:
            extype, value, tb = sys.exc_info()
            sys.last_type = extype
            sys.last_value = value
            sys.last_traceback = tb
            tblist = traceback.extract_tb(tb)
            del tblist[:1]
            lst = traceback.format_list(tblist)
            if lst:
                lst.insert(0, "Traceback (most recent call last):\n")
            exc_msg = traceback.format_exception_only(extype, value)
            lst[len(lst):] = exc_msg
        finally:
            tblist = tb = None

        if self.filename.startswith("<literate "):
            self._fakeout.write(''.join(lst))
        else:
            if sys.excepthook is sys.__excepthook__:
                self.write(''.join(lst))
            else:
                sys.excepthook(extype, value, tb)

        self.exc_msg = ''.join(exc_msg)

    def _run_code(self, chunk, chunk_ind, pause=True):
        self.filename = "<literate {name}[{num}]>".format(name=self.name,
                                                          num=chunk_ind)
        more = False
        res = ""
        lines = chunk.source.split("\n")
        for line in lines[:-1]:
            res, exc, more = self._process_code_line(line, res, more)

        if more:
            res, exc, more = self._process_code_line(lines[-1], res, more)

        if self.use_ansi:
            mgr = ansi.BoxMaker(self)
        else:
            mgr = noop_mgr(self)

        if chunk.want is not None or chunk.exc_msg is not None:
            if len(res) > 0 or exc is not None:
                # compare to the expected output
                # TODO(sross): convert options to optionsflags
                optionsflags = 0

                checker = self._output_checker
                if exc is None:
                    same = self._output_checker.check_output(chunk.want,
                                                             res, 0)
                    if not same:
                        self.write('\n')
                        with mgr as maker:
                            maker.write('Warning, output different from '
                                        'expected:\n')
                            maker.write('================================'
                                        '========\n\n')
                            diff = checker.output_difference(chunk, res,
                                                             optionsflags)
                            maker.write(diff)
                        self.write('\n')
                    else:
                        self.write(res)
                elif chunk.exc_msg is None:
                    self.write('\n')
                    with mgr as maker:
                        maker.write('Warning, unexpected exception:\n')
                        maker.write('==============================\n\n')
                        maker.write(exc)
                    self.write('\n')
                else:
                    same_ex = checker.check_output(chunk.exc_msg,
                                                   exc, optionsflags)
                    if not same_ex:
                        self.write('\n')
                        with mgr as maker:
                            maker.write('Warning, exception different from '
                                        'expected:\n')
                            maker.write('=================================='
                                        '=========\n\n')
                            diff = checker.output_difference(chunk, res,
                                                             optionsflags)
                            maker.write(diff)
                        self.write('\n')
                    else:
                        self.write(res)
        else:
            if exc is not None:
                self.write('\n')
                with mgr as maker:
                    maker.write('Warning, unexpected exception:\n')
                    maker.write('==============================\n\n')
                    maker.write(exc)
                self.write('\n')
            else:
                self.write(res)

        if not self.pause:
            self.write(sys.ps1 + '\n')

    def no_echo_input(self, prompt):
        with warnings.catch_warnings():
            res = getpass.getpass(prompt)
        return res

    def interact(self, lit_string, name, pause=True, interactive=True):
        self.name = name
        self.pause = pause
        self.interactive = interactive

        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = ">>> "

        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = "... "

        try:
            sys.ps3
        except AttributeError:
            sys.ps3 = '>>> '

        if self._env_driver is not None:
            extra_locals = self._env_driver.setup()
            self.locals.update(extra_locals)
            extra_banner = self._env_driver.banner
            driver_text = " ({0})".format(self._env_driver.DRIVER_NAME)
        else:
            extra_banner = ""
            driver_text = ""

        cprt = ('Type "help", "copyright", "credits" or "license" for '
                'more information about Python.')
        self.write("Literate Python Shell{driver_text}\nPython {ver} "
                   "on {platform}\n{cprt}\n"
                   "{extra_banner}\n\n".format(driver_text=driver_text,
                                               ver=sys.version,
                                               platform=sys.platform,
                                               cprt=cprt,
                                               extra_banner=extra_banner))

        if not interactive and pause:
            self.write('Press enter to continue after a code block\n\n')

        try:
            parser = self.code_parser
            start = True
            self.chunks = list(parser.parse(lit_string, name))
            for chunk_ind, chunk in enumerate(self.chunks):
                if isinstance(chunk, parsers.CodeChunk):
                    self._run_code(chunk, chunk_ind)
                elif not chunk:
                    continue
                else:
                    if not start and pause and interactive:
                        self.filename = "<stdin>"
                        more = False
                        blanks = 0
                        while blanks < 2:
                            blank, more = self._interact_once(more)

                            if blank:
                                blanks += 1
                            else:
                                blanks = 0

                            if more is None:
                                return

                        # reset exc_msg so it doesn't get
                        # raised after the next code block
                        self.exc_msg = None
                    elif not start and pause:
                        self.no_echo_input(sys.ps3)

                    self.write(self.text_formatter.format(chunk))

                start = False

            complete_msg = ("\n{file} complete! Continuing to interactive "
                            "console...\n\n".format(file=self.name))

            if self.use_ansi:
                self.write(ansi.with_codes(complete_msg, 1))
            else:
                self.write(complete_msg)

            self.filename = "<stdin>"
            self.locals['con'] = self
            more = False
            while more is not None:
                blank, more = self._interact_once(more)
        finally:
            if self._env_driver is not None:
                self._env_driver.teardown()



# Copyright 2014, Solly Ross (see LICENSE.txt)
import six


_ESC = '\x1b'
_CSI = _ESC + '['
_SET_FONT_G1 = _ESC + ')'
_SAVE_STATE = _ESC + '7'
_LOAD_STATE = _ESC + '8'

_SHIFT_OUT = '\x0E'
_SHIFT_IN = '\x0F'

_GRAPHICS_FONT = '0'


def with_codes(text, *codes):
    str_codes = [six.text_type(code) for code in codes]
    return _CSI + (';'.join(str_codes)) + 'm' + text + _CSI + '0m'


class BoxMaker(object):
    def __init__(self, to, padding=1):
        self.output = to
        self.padding = padding

    def __enter__(self):
        self.curr_len = 0
        self.width = 0
        self.buff = ['']
        return self

    def _output_line(self, line):
        self.output.write(_SHIFT_OUT + 'x' + _SHIFT_IN)
        self.output.write(' ' * self.padding)
        self.output.write(line)
        if len(line) < self.unpadded_width:
            self.output.write(' ' * (self.unpadded_width - len(line)))
        self.output.write(' ' * self.padding)
        self.output.write(_SHIFT_OUT + 'x' + _SHIFT_IN + '\n')

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.output.write(_SAVE_STATE)
        self.output.write(_SET_FONT_G1 + _GRAPHICS_FONT)

        self.unpadded_width = self.width
        self.width += self.padding * 2

        # suppress a final extraneous newline
        if self.buff[-1] == '':
            del self.buff[-1]

        # top
        self.output.write(_SHIFT_OUT + 'l' + ('q' * self.width) + 'k' +
                          _SHIFT_IN + '\n')

        for i in six.moves.range(self.padding):
            self._output_line('')

        for line in self.buff:
            self._output_line(line)

        for i in six.moves.range(self.padding):
            self._output_line('')

        # bottom
        self.output.write(_SHIFT_OUT + 'm' + ('q' * self.width) + 'j' +
                          _SHIFT_IN + '\n')
        self.output.write(_LOAD_STATE)

        return False

    def write(self, contents):
        lines = contents.split('\n')
        if len(lines) > 1:
            self.buff[-1] += lines[0]
            self.curr_len += len(lines[0])
            if self.curr_len > self.width:
                self.width = self.curr_len

            for line in lines[1:]:
                self.curr_len = len(line)
                if self.curr_len > self.width:
                    self.width = self.curr_len

                self.curr_len = 0

            self.buff.extend(lines[1:])
        else:
            # no newline
            self.buff[-1] += lines[0]
            self.curr_len += len(lines[0])



#! /usr/bin/env python
# Copyright 2014, Solly Ross (see LICENSE.txt)
from __future__ import print_function

import argparse
import os.path
import sys

import pkg_resources as pkgres

from yalpt import core


parser = argparse.ArgumentParser()
parser.formatter_class = argparse.RawDescriptionHelpFormatter
parser.usage = "%(prog)s [OPTION]... file"
parser.description = "Yet Another Literate Python Tool"
parser.epilog = """

YAPLT is a tool for interacting with documentation
interspersed with python code.  Unlike other literate
python tools, most of which are designed for producing
static documentation, YALPT is designed for writing
interactive tutorials.

By default, YALPT will drop into an interactive session
after each code block, allowing you to experiment around
with the results of the last block of code.  Once you are
done experimenting, simply enter two blank lines,
and YALPT will continue with the file.

Python code should be written so that it would be detectable
by the doctest module.  This means that code samples should
look like interactive python sessions.

Other than that, the format can vary.  You are free to use
your own style for the documentation, although YALPT comes
with a built-in formatter that renders a subset of Markdown
with ANSI escape codes, so it looks nice in your terminal.
"""

parser.add_argument('file',
                    help='The YALPT literate python file to run')
parser.add_argument('--no-pause', action='store_false', dest='pause',
                    default=True,
                    help="Don't pause after every code block.  "
                         "Implies --no-interactive")
parser.add_argument('--no-interactive', action='store_false',
                    dest='interactive', default=True,
                    help="Don't drop to an interactive console "
                         "after every code block")
parser.add_argument('-f', '--format', default=None,
                    help="Format the text portions using ANSI "
                         "escape codes.  Currently accepts 'md' and 'none'. "
                         "YALPT will attempt to automatically set this based "
                         "on file extension.")
parser.add_argument('-p', '--code-parser', default='doctest',
                    help="Which method was used to embed code samples in the "
                         "document.  Currently accepts 'doctest' and "
                         "'markdown'.")
parser.add_argument('--no-readline', dest='readline', action='store_false',
                    default=True,
                    help="Don't import readline for completion and history")
parser.add_argument('--no-ansi', dest='ansi', action='store_false',
                    default=True,
                    help="Don't use ANSI escape codes to provide color and "
                         "formatting.  Will not work with --format.")
parser.add_argument('-e, --env-driver', dest='env_driver', default=None,
                    help="Use the given env driver to set up the environment"
                         "in which the code executes.  Use [] to pass a "
                         "parameter.")

args = parser.parse_args()

filename = os.path.basename(args.file)

# code parser
parsers = pkgres.iter_entry_points('yalpt.parsers', args.code_parser)

try:
    code_parser_cls_loader = next(parsers)
except StopIteration:
    sys.exit("Could not load code parser %s" % args.code_parser)


code_parser = code_parser_cls_loader.load()()

# text formatter
if not args.ansi and args.format:
    sys.exit("Cannot use a formatter without ANSI escape code support!")

if args.format is None:
    if '.' in filename:
        args.format = filename.rsplit('.', 1)[-1]
    else:
        args.format = 'none'

if not args.ansi:
    args.format = 'none'

formatters = pkgres.iter_entry_points('yalpt.formatters', args.format)

try:
    formatter_cls_loader = next(formatters)
except StopIteration:
    print("Warning: cannot load formatter for the .%s extension -- "
          "no formatters found." % args.format, file=sys.stderr)

    formatter_cls_loader = next(pkgres.iter_entry_points('yalpt.formatters',
                                                         'none'))

text_formatter = formatter_cls_loader.load()()


env_driver = None
if args.env_driver:
    if '[' in args.env_driver:
        bracket_pos = args.env_driver.index('[')
        env_driver_name = args.env_driver[:bracket_pos]
        env_driver_args = [args.env_driver[bracket_pos + 1:-1]]
    else:
        env_driver_name = args.env_driver
        env_driver_args = []

    env_drivers = pkgres.iter_entry_points('yalpt.env_drivers',
                                           args.env_driver)
    try:
        env_driver_loader = next(env_drivers)
        env_driver = env_driver_loader.load()(*env_driver_args)
    except StopIteration:
        sys.exit("Error: cannot load env driver %s -- no such "
                 "env driver found" % args.env_driver)
        env_driver = None

interpreter = core.LiterateInterpreter(text_formatter=text_formatter,
                                       code_parser=code_parser,
                                       use_ansi=args.ansi,
                                       use_readline=args.readline,
                                       env_driver=env_driver)
with open(args.file) as f:
    interpreter.interact(f.read(), filename,
                         pause=args.pause, interactive=args.interactive)



from setuptools import setup  # noqa

setup(name='yalpt',
      version='1.0.0',
      description='Yet Another Literate Python Tool',
      long_description=open('README.md').read(),
      scripts=['run-lit.py'],
      author='Solly Ross',
      author_email='sross@redhat.com',
      license='ISC and PSF',
      url='https://github.com/directxman12/yalpt',
      packages=['yalpt'],
      install_requires=['six'],
      keywords='literate documentation tutorial',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Topic :: Documentation',
          'Topic :: Software Development',
          'License :: OSI Approved :: ISC License (ISCL)',
          'License :: OSI Approved :: Python Software Foundation License',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
      ],
      entry_points={
          'yalpt.formatters': [
              'md = yalpt.formatters:MarkdownFormatter',
              'none = yalpt.formatters:NoopFormatter',
          ],
          'yalpt.parsers': [
              'doctest = yalpt.parsers:DocTestParser',
              'markdown = yalpt.parsers:MarkdownParser',
          ]
      })



# Copyright 2014, Solly Ross (see LICENSE.txt)
import re

import six

from yalpt import ansi_helper as ansi


class NoopFormatter(object):
    def format(self, s):
        return s


class MarkdownFormatter(object):
    FORMAT_CODES = {'\x01': 1,
                    '\x04': 3,
                    '\x05': 7,
                    '\x1C': [48, 5, 0xea],
                    '\x1D': ([4, 1], [48, 5, 0xea]),
                    '\x1E': ([], [48, 5, 0xea]),
                    '\x06': 4,
                    '\x07': 9}

    FORMAT_CHARS = {'\x01': u'**',
                    '\x04': u'*',
                    '\x05': u'==',
                    '\x06': u'_',
                    '\x1C': u' ',  # use spaces instead of '`'
                    '\x1D': (u'', u'    '),  # use spaces instead of '`'
                    '\x1E': (u'', u'    '),
                    '\x07': u'~~'}

    # BOLD MUST BE MATCHED FIRST, AND SUBSTITUTED OUT
    # TODO(sross): this will miss the case of *a*, etc

    # bold: **text** --> \x01
    BOLD_RE = re.compile(r'(?<!\\)\*\*(?P<text>[^ *].*?[^ \\])\*\*')

    # italics: *text* --> \x04
    ITALICS_RE = re.compile(r'(?<!\\)\*(?P<text>[^ ].*?[^ \\])\*')

    # highlight: ==text== --> \x05
    HIGHLIGHT_RE = re.compile(r'(?<!\\)==(?P<text>[^ =].*?[^ \\])==')

    # underline: _text_ --> \x06
    UNDERLINE_RE = re.compile(r'(?<!\\|\w)_(?P<text>[^ ].*?[^ \\])_')

    # strikethrough: ~~text~~ --> \x07
    STRIKETHROUGH_RE = re.compile(r'(?<!\\)~~(?P<text>[^ ].*?[^ \\])~~')

    # code: ```language\ncontents\n```
    FENCED_RE = re.compile(
        r'```(?P<language>[^' + '\n' + r']+?)?\n(?P<text>.+?)\n```', re.DOTALL)

    INDENTED_RE = re.compile(r'(\n^    .+$)+', re.MULTILINE)

    # code: `text`
    CODE_RE = re.compile(r'`(?P<text>.+?)`')

    AFTER_RE = re.compile('\x02(?P<code>.)(?P<text>.+?)\x03')

    HEADER_RE1 = re.compile(r'^(?P<start>#+) (?P<name>.+?)(?P<end> #+)?$',
                            re.MULTILINE)
    HEADER_RE2 = re.compile(r'^(?P<text>.+)\n(?P<underline>(==+)|(--+))$',
                            re.MULTILINE)

    def _add_ansi_formatting(self, match):
        sub_code = match.group('code')
        multiline = False
        prefix = None

        if sub_code in ('\x1C', '\x1E'):
            # deal with the special code blocks thing
            block_num = int(match.group('text'))
            inner_text = self.code_blocks[block_num]
            if sub_code == '\x1E':
                multiline = True
                prefix = ''
        elif sub_code == '\x1D':
            block_num = int(match.group('text'))
            prefix, inner_text = self.code_blocks[block_num]
            multiline = True
        else:
            inner_text = match.group('text')

        if multiline:
            prefix_codes, codes = self.FORMAT_CODES[sub_code]
            if isinstance(prefix_codes, int):
                prefix_codes = [prefix_codes]
        else:
            codes = self.FORMAT_CODES[sub_code]

        if isinstance(codes, int):
            codes = [codes]

        if multiline:
            chars, line_chars = self.FORMAT_CHARS[sub_code]
        else:
            chars = self.FORMAT_CHARS[sub_code]

        text = chars + inner_text + chars

        if multiline:
            raw_lines = text.splitlines()
            max_len = max(len(line) for line in raw_lines)

            res_lines = [("{1}{0: <%s}{1}" % max_len).format(line, line_chars)
                         for line in raw_lines]

            res = '\n'.join(ansi.with_codes(line, *codes)
                            for line in res_lines)
        else:
            res = ansi.with_codes(text, *codes)

        if multiline and prefix is not None:
            prefix_raw = ('{1}{0: <%s}{1}' % max_len).format(prefix,
                                                             line_chars)
            res = '%s\n%s' % (ansi.with_codes(prefix_raw, *prefix_codes), res)

        return res

    def _remove_code_block(self, match):
        self.code_blocks.append((match.group('language'), match.group('text')))
        return '\x02\x1D' + six.text_type(len(self.code_blocks)-1) + '\x03'

    def _remove_code_inline(self, match):
        self.code_blocks.append(match.group('text'))
        return '\x02\x1C' + six.text_type(len(self.code_blocks)-1) + '\x03'

    def _remove_code_indented(self, match):
        self.code_blocks.append(
            '\n'.join(line[4:] for line in match.group(0).splitlines()[1:]))
        return '\x02\x1E' + six.text_type(len(self.code_blocks)-1) + '\x03'

    def format(self, s):
        self.code_blocks = []

        # first, temporarily remove any code blocks,
        # as we don't want them formatted
        s = self.FENCED_RE.sub(self._remove_code_block, s)
        s = self.CODE_RE.sub(self._remove_code_inline, s)
        s = self.INDENTED_RE.sub(self._remove_code_indented, s)

        # first find bold, and then italic, then the rest
        s = self.BOLD_RE.sub('\x02\x01\\g<text>\x03', s)
        s = self.ITALICS_RE.sub('\x02\x04\\g<text>\x03', s)
        s = self.HIGHLIGHT_RE.sub('\x02\x05\\g<text>\x03', s)
        s = self.UNDERLINE_RE.sub('\x02\x06\\g<text>\x03', s)
        s = self.STRIKETHROUGH_RE.sub('\x02\x07\\g<text>\x03', s)

        # Find and format headers
        s = self.HEADER_RE1.sub(ansi.with_codes(r'\g<start> \g<name> \g<end>',
                                1), s)
        s = self.HEADER_RE2.sub(ansi.with_codes('\\g<text>\n\\g<underline>',
                                1), s)

        # now return everything to normal, with the ANSI encoding
        s = self.AFTER_RE.sub(self._add_ansi_formatting, s)

        return s
